import java.io.File;
import java.io.IOException;
import java.util.HashMap;
import java.util.StringTokenizer;

import javax.xml.namespace.NamespaceContext;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpression;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.w3c.dom.Attr;
import org.w3c.dom.Comment;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.xml.sax.SAXException;

/**
 * 
 */

/**
 * @author gosiaw
 * 
 */
public class DEBB2DCWoRMSTranslator {

	private static final String DEFAULT_OUTPUT_FILE_NAME = "xml/DCWORMS_TEST.xml";
	private static final String DEFAULT_INPUT_FILE_NAME = "xml/PLMXML_HLRSSmallServerRoom.xml";
//	private static final String DEFAULT_INPUT_FILE_NAME = "xml/PLMXML_HLRSRECSv2.xml";
	private static final String DEFAULT_DEBB_COMPONENT_FILE_NAME = "xml/DEBBComponents.xml";
	private static DocumentBuilder xmlDocumentBuilder;
	private static DocumentBuilderFactory domFactory;

	private static HashMap<String, String> computingResourceClasses = null;
	
	/**
	 * 
	 */
	public DEBB2DCWoRMSTranslator() {
		// TODO Auto-generated constructor stub
	}

	/**
	 * @param args
	 */
	public static void main(String[] args) {
		String inputFileName = DEFAULT_INPUT_FILE_NAME;
		String outputFileName = DEFAULT_OUTPUT_FILE_NAME;

		if (args != null && args.length > 0) {
			inputFileName = args[0];
			if (inputFileName != null && inputFileName.isEmpty() == false) {
				inputFileName = DEFAULT_INPUT_FILE_NAME;
			}

			outputFileName = args[1];
			if (outputFileName != null && outputFileName.isEmpty() == false) {
				outputFileName = DEFAULT_OUTPUT_FILE_NAME;
			}
		}

		try {
			initialize();

			// Open and read content of the input file (i.e. PLM XML document)
			Document plmxmlDocument = readInputFile(inputFileName);

			XPathFactory xpathFactory = XPathFactory.newInstance();
			XPath xpath = xpathFactory.newXPath();
			NamespaceContext nsContext = new DEBBNamespaceContext();
			xpath.setNamespaceContext(nsContext);

			// Find root component (top level DEBB) id
			XPathExpression expr = xpath
					.compile("//plm:ProductDef/plm:InstanceGraph/@rootRefs");
			Object result = expr.evaluate(plmxmlDocument, XPathConstants.NODE);
			Node rootId = (Node) result;
			System.out.println(rootId.getNodeName());
			System.out.println(rootId.getNodeValue());

			// Creating destination, dcworms document
			Document dcwormsDocument = xmlDocumentBuilder.newDocument();
			// Creating root, environment, element
			Element rootElement = dcwormsDocument.createElement("environment");
			dcwormsDocument.appendChild(rootElement);
			Comment comment = dcwormsDocument.createComment("CoolEmAll testbed description");
			rootElement.appendChild(comment);
			
			// Creating resources element
			Element resourcesElement = dcwormsDocument.createElement("resources");
			rootElement.appendChild(resourcesElement);
			
			// Creating (main) scheduler element
			// TODO: Fill scheduler with reasonable data!
			Element schedulerElement = dcwormsDocument.createElement("scheduler");
			comment = dcwormsDocument.createComment("Default scheduler");
			schedulerElement.appendChild(comment);
			resourcesElement.appendChild(schedulerElement);

			// Transform top level DEBB and all nested DEBBs
			transformDEBB(plmxmlDocument, dcwormsDocument, resourcesElement,
					xpath, rootId.getNodeValue());

			// write the content into xml file
			saveOutputFile(outputFileName, dcwormsDocument);

		} catch (ParserConfigurationException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (SAXException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (XPathExpressionException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

	}

	private static void initialize() throws ParserConfigurationException {
		domFactory = DocumentBuilderFactory.newInstance();
		domFactory.setNamespaceAware(true); // never forget this!
		domFactory.setValidating(true);
		domFactory.setAttribute(
				"http://java.sun.com/xml/jaxp/properties/schemaLanguage",
				"http://www.w3.org/2001/XMLSchema");

		xmlDocumentBuilder = domFactory.newDocumentBuilder();
		
		//	TODO: Check it this is correct!
		//	TODO: Move it to configuration file
		computingResourceClasses = new HashMap<String, String>();
		computingResourceClasses.put("ComputeBox2", "ComputeBox2");
		computingResourceClasses.put("ComputeBox1", "ComputeBox1");
		computingResourceClasses.put("NodeGroup", "NodeGroup");
		computingResourceClasses.put("Node", "Node");

	}

	private static Document readInputFile(String inputFileName) {
		Document plmxmlDocument;
		try {
			plmxmlDocument = xmlDocumentBuilder.parse(inputFileName);
			return plmxmlDocument;
		} catch (SAXException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		return null;
	}

	private static void saveOutputFile(String outputFileName,
			Document dcwormsDocument) {

		try {
			TransformerFactory transformerFactory = TransformerFactory
					.newInstance();
			Transformer transformer;

			transformer = transformerFactory.newTransformer();

			DOMSource source = new DOMSource(dcwormsDocument);
			StreamResult transformResult = new StreamResult(new File(
					outputFileName));

			// Output to console for testing
			// StreamResult result = new StreamResult(System.out);

			transformer.transform(source, transformResult);

			System.out.println("File " + outputFileName + " saved!");

		} catch (TransformerConfigurationException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (TransformerException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

	}

	/**
	 * Translates specified DEBB and all its nested sub DEBBS
	 * 
	 * @param plmxmlDocument
	 * @param dcwormsDocument
	 *            TODO
	 * @param xpath
	 * @param instanceId
	 * @return
	 * @throws Exception
	 */
	public static void transformDEBB(Document plmxmlDocument,
			Document dcwormsDocument, Element debbParentElement, XPath xpath,
			String instanceId) throws Exception {

		if (instanceId == null) {
			throw new Exception("Null DEBB identifier");
		}
		if (plmxmlDocument == null) {
			throw new Exception("Null PLM XML document");
		}

		if (debbParentElement == null) {
			throw new Exception("Null parent element in dcworms document");
		}

		// Find DEBB with given id and transform it
		// 1) ProductInstance -
		XPathExpression expr = xpath
				.compile("//plm:ProductDef/plm:InstanceGraph/plm:ProductInstance[@id='"
						+ instanceId + "']");
		Object result = expr.evaluate(plmxmlDocument, XPathConstants.NODE);
		Node productInstance = (Node) result;
		Node debbName = productInstance.getAttributes().getNamedItem("name");
		System.out.println("name=" + debbName.getNodeValue());
		Node partRef = productInstance.getAttributes().getNamedItem("partRef");
		System.out.println("partRef=" + partRef.getNodeValue());
		String partId = partRef.getNodeValue().substring(1); // Get rid of # at
																// the beginning

		// 2) ProductRevisionView
		expr = xpath
				.compile("//plm:ProductDef/plm:InstanceGraph/plm:ProductRevisionView[@id='"
						+ partId + "']");
		result = expr.evaluate(plmxmlDocument, XPathConstants.NODE);
		Node productRevisionView = (Node) result;
		debbName = productRevisionView.getAttributes().getNamedItem("name");
		System.out.println("name=" + debbName.getNodeValue());

		// 3) UserValue (to extract DEBB name, level or DEBBComponent file)
		expr = xpath
				.compile("plm:UserData/plm:UserValue[@title='DEBBLevel']/@value");
		result = expr.evaluate(productRevisionView, XPathConstants.NODE);
		Node debbLevel = (Node) result;
		if (debbLevel != null) {
			System.out.println("debbLevel=" + debbLevel.getNodeValue());
		}
		expr = xpath.compile("plm:UserData/plm:UserValue[@title='DEBBComponentFile']/@value");
		result = expr.evaluate(productRevisionView, XPathConstants.NODE);
		Node debbComponentFileName = (Node)result;
		if (debbComponentFileName == null) {
			//	Default DEBBComponent file is assumed
			
			
		}
		System.out.println("debbComponentFileName=" + debbComponentFileName);

		// Add computingResource element for the DEBB
		if (debbLevel != null) {
			Element computingResourceElement = dcwormsDocument
					.createElement("computingResource");
			Attr nameAttribute = dcwormsDocument.createAttribute("name");
			nameAttribute.setValue(debbName.getNodeValue());
			computingResourceElement.setAttributeNode(nameAttribute);
			// TODO: how to translate DEBB level to class attribute?
			Attr classAttribute = dcwormsDocument.createAttribute("class");
			String dcwormsClass = computingResourceClasses.get(debbLevel.getNodeValue());
			classAttribute.setValue(dcwormsClass);
			computingResourceElement.setAttributeNode(classAttribute);
			debbParentElement.appendChild(computingResourceElement);

			// Get nested instances - nested DEBBs and ask them to transform
			// themselves
			Node instanceRefs = productRevisionView.getAttributes()
					.getNamedItem("instanceRefs");
			// Check if anything is nested at all
			if (instanceRefs != null) {
				String instances = instanceRefs.getNodeValue();
				System.out.println("instances=" + instances);
				// Separate instance ids
				StringTokenizer tokenizer = new StringTokenizer(instances, " ");
				while (tokenizer.hasMoreTokens()) {
					String instance = tokenizer.nextToken();
					System.out.println(instance);
					// Go down in the hierarchy
					transformDEBB(plmxmlDocument, dcwormsDocument,
							computingResourceElement, xpath, instance);
					System.out.println();
				}
			}
		}
	}

}
